"""CassandraClient entity."""
from ssl import CERT_REQUIRED, PROTOCOL_TLSv1
from typing import Dict, List, Optional

from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import Cluster, ResponseFuture, Session
from cassandra.policies import RoundRobinPolicy
from cassandra.query import dict_factory
from typing_extensions import TypedDict

from butterfree.clients import AbstractClient


class CassandraColumn(TypedDict):
    """Type for cassandra columns.

    It's just a type abstraction, we can use it or a normal dict

    >>> def function(column: CassandraColumn) -> CassandraColumn:
    ...     return column
    >>> # The following two lines will pass in the type checking
    >>> function({'column_name': 'test', 'type': 'integer', 'primary_key': False})
    >>> function(CassandraColumn(column_name='test', type='integer', primary_key=False))

    """

    column_name: str
    type: str
    primary_key: bool


class CassandraClient(AbstractClient):
    """Cassandra Client.

    Attributes:
        user: username to use in connection.
        password: password to use in connection.
        keyspace: key space used in connection.
        host: cassandra endpoint used in connection.
    """

    def __init__(
        self,
        host: List[str],
        keyspace: str,
        user: Optional[str] = None,
        password: Optional[str] = None,
    ) -> None:
        self.host = host
        self.keyspace = keyspace
        self.user = user
        self.password = password
        self._session: Optional[Session] = None

    @property
    def conn(self, *, ssl_path: str = None) -> Session:  # type: ignore
        """Establishes a Cassandra connection."""
        auth_provider = (
            PlainTextAuthProvider(username=self.user, password=self.password)
            if self.user is not None
            else None
        )
        ssl_opts = (
            {
                "ca_certs": ssl_path,
                "ssl_version": PROTOCOL_TLSv1,
                "cert_reqs": CERT_REQUIRED,
            }
            if ssl_path is not None
            else None
        )

        cluster = Cluster(
            contact_points=self.host,
            auth_provider=auth_provider,
            ssl_options=ssl_opts,
            load_balancing_policy=RoundRobinPolicy(),
        )
        self._session = cluster.connect(self.keyspace)
        self._session.row_factory = dict_factory
        return self._session

    def sql(self, query: str) -> ResponseFuture:
        """Executes desired query.

        Attributes:
            query: desired query.

        """
        if not self._session:
            raise RuntimeError("There's no session available for this query.")
        return self._session.execute(query)

    def get_schema(self, table: str) -> List[Dict[str, str]]:
        """Returns desired table schema.

        Attributes:
            table: desired table.

        Returns:
            A list dictionaries in the format
            [{"column_name": "example1", type: "cql_type"}, ...]

        """
        query = (
            f"SELECT column_name, type FROM system_schema.columns "  # noqa
            f"WHERE keyspace_name = '{self.keyspace}' "  # noqa
            f"  AND table_name = '{table}';"  # noqa
        )

        response = list(self.sql(query))

        if not response:
            raise RuntimeError(
                f"No columns found for table: {table}" f"in key space: {self.keyspace}"
            )

        return response

    def _get_create_table_query(
        self, columns: List[CassandraColumn], table: str,
    ) -> str:
        """Creates CQL statement to create a table."""
        parsed_columns = []
        primary_keys = []

        for col in columns:
            col_str = f"{col['column_name']} {col['type']}"
            if col["primary_key"]:
                primary_keys.append(col["column_name"])
            parsed_columns.append(col_str)

        joined_parsed_columns = ", ".join(parsed_columns)

        if len(primary_keys) > 0:
            joined_primary_keys = ", ".join(primary_keys)
            columns_str = (
                f"{joined_parsed_columns}, PRIMARY KEY ({joined_primary_keys})"
            )
        else:
            columns_str = joined_parsed_columns

        query = f"CREATE TABLE {self.keyspace}.{table} " f"({columns_str}); "

        return query

    def create_table(self, columns: List[CassandraColumn], table: str) -> None:
        """Creates a table.

        Attributes:
            columns: a list dictionaries in the format
                [{"column_name": "example1", type: "cql_type", primary_key: True}, ...]
        """
        query = self._get_create_table_query(columns, table)

        self.sql(query)



"""Migration entity."""
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Dict, List, Set

from butterfree.clients import AbstractClient
from butterfree.load.writers.writer import Writer
from butterfree.transform import FeatureSet


@dataclass
class Diff:
    """DataClass to help identifying different types of diff between schemas."""

    class Kind(Enum):
        """Mapping actions to take given a difference between columns of a schema."""

        ADD = auto()
        ALTER_KEY = auto()
        ALTER_TYPE = auto()
        DROP = auto()

    column: str
    kind: Kind
    value: Any

    def __hash__(self) -> int:
        return hash((self.column, self.kind, self.value))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            raise NotImplementedError
        return (
            self.column == other.column
            and self.kind == other.kind
            and self.value == other.value
        )


class DatabaseMigration(ABC):
    """Abstract base class for Migrations."""

    def __init__(self, client: AbstractClient) -> None:
        self._client = client

    @abstractmethod
    def _get_create_table_query(
        self, columns: List[Dict[str, Any]], table_name: str
    ) -> Any:
        """Creates desired statement to create a table.

        Args:
            columns: object that contains column's schemas.
            table_name: table name.

        Returns:
            Create table query.

        """
        pass

    @abstractmethod
    def _get_alter_table_add_query(self, columns: List[Diff], table_name: str) -> str:
        """Creates desired statement to add columns to a table.

        Args:
            columns: list of Diff objects with ADD kind.
            table_name: table name.

        Returns:
            Alter table query.

        """
        pass

    @abstractmethod
    def _get_alter_table_drop_query(self, columns: List[Diff], table_name: str) -> str:
        """Creates desired statement to drop columns from a table.

        Args:
            columns: list of Diff objects with DROP kind.
            table_name: table name.

        Returns:
            Drop columns from a given table query.

        """
        pass

    @abstractmethod
    def _get_alter_column_type_query(self, columns: List[Diff], table_name: str) -> str:
        """Creates desired statement to alter columns' types.

        Args:
            columns: list of Diff objects with ALTER_TYPE kind.
            table_name: table name.

        Returns:
            Alter column type query.

        """
        pass

    def _get_queries(
        self, schema_diff: Set[Diff], table_name: str, write_on_entity: bool = None
    ) -> Any:
        """Create the desired queries for migration.

        Args:
            schema_diff: list of Diff objects.
            table_name: table name.

        Returns:
            List of queries.

        """
        add_items = []
        drop_items = []
        alter_type_items = []
        alter_key_items = []

        for diff in schema_diff:
            if diff.kind == Diff.Kind.ADD:
                add_items.append(diff)
            elif diff.kind == Diff.Kind.ALTER_TYPE:
                alter_type_items.append(diff)
            elif diff.kind == Diff.Kind.DROP:
                drop_items.append(diff)
            elif diff.kind == Diff.Kind.ALTER_KEY:
                alter_key_items.append(diff)

        queries = []
        if add_items:
            alter_table_add_query = self._get_alter_table_add_query(
                add_items, table_name
            )
            queries.append(alter_table_add_query)
        if drop_items:
            if write_on_entity:
                logging.info(
                    "Features will not be dropped automatically "
                    "when data is loaded to an entity table"
                )
            else:
                drop_columns_query = self._get_alter_table_drop_query(
                    drop_items, table_name
                )
                queries.append(drop_columns_query)
        if alter_type_items:
            alter_column_types_query = self._get_alter_column_type_query(
                alter_type_items, table_name
            )
            queries.append(alter_column_types_query)
        if alter_key_items:
            logging.info("This operation is not supported by Spark.")

        return queries

    def create_query(
        self,
        fs_schema: List[Dict[str, Any]],
        table_name: str,
        db_schema: List[Dict[str, Any]] = None,
        write_on_entity: bool = None,
    ) -> Any:
        """Create a query regarding a data source.

        Returns:
            The desired queries for the given database.

        """
        if not db_schema:
            return [self._get_create_table_query(fs_schema, table_name)]

        schema_diff = self._get_diff(fs_schema, db_schema)

        return self._get_queries(schema_diff, table_name, write_on_entity)

    @staticmethod
    def _get_diff(
        fs_schema: List[Dict[str, Any]], db_schema: List[Dict[str, Any]],
    ) -> Set[Diff]:
        """Gets schema difference between feature set and the table of a given db.

        Args:
            fs_schema: object that contains feature set's schemas.
            db_schema: object that contains the table of a given db schema.

        Returns:
            Object with schema differences.

        """
        db_columns = set(item.get("column_name") for item in db_schema)
        fs_columns = set(item.get("column_name") for item in fs_schema)

        add_columns = fs_columns - db_columns
        drop_columns = db_columns - fs_columns

        # This could be way easier to write (and to read) if the schemas were a simple
        # Dict[str, Any] where each key would be the column name itself...
        # but changing that could break things so:
        # TODO version 2 change get schema to return a dict(columns, properties)
        add_type_columns = dict()
        alter_type_columns = dict()
        alter_key_columns = dict()
        for fs_item in fs_schema:
            if fs_item.get("column_name") in add_columns:
                add_type_columns.update(
                    {fs_item.get("column_name"): fs_item.get("type")}
                )
            for db_item in db_schema:
                if fs_item.get("column_name") == db_item.get("column_name"):
                    if fs_item.get("type") != db_item.get("type"):
                        alter_type_columns.update(
                            {fs_item.get("column_name"): fs_item.get("type")}
                        )
                    if fs_item.get("primary_key") != db_item.get("primary_key"):
                        alter_key_columns.update(
                            {fs_item.get("column_name"): fs_item.get("primary_key")}
                        )
                    break

        schema_diff = set(
            Diff(str(col), kind=Diff.Kind.ADD, value=value)
            for col, value in add_type_columns.items()
        )
        schema_diff |= set(
            Diff(str(col), kind=Diff.Kind.DROP, value=None) for col in drop_columns
        )
        schema_diff |= set(
            Diff(str(col), kind=Diff.Kind.ALTER_TYPE, value=value)
            for col, value in alter_type_columns.items()
        )
        schema_diff |= set(
            Diff(str(col), kind=Diff.Kind.ALTER_KEY, value=None)
            for col, value in alter_key_columns.items()
        )
        return schema_diff

    def _get_schema(self, table_name: str) -> List[Dict[str, Any]]:
        """Get a table schema in the respective database.

        Args:
            table_name: Table name to get schema.

        Returns:
            Schema object.
        """
        try:
            db_schema = self._client.get_schema(table_name)
        except Exception:  # noqa
            db_schema = []
        return db_schema

    def apply_migration(self, feature_set: FeatureSet, writer: Writer,) -> None:
        """Apply the migration in the respective database.

        Args:
            feature_set: the feature set.
            writer: the writer being used to load the feature set.
        """
        logging.info(f"Migrating feature set: {feature_set.name}")

        table_name = (
            feature_set.name if not writer.write_to_entity else feature_set.entity
        )

        fs_schema = writer.db_config.translate(feature_set.get_schema())
        db_schema = self._get_schema(table_name)

        queries = self.create_query(
            fs_schema, table_name, db_schema, writer.write_to_entity
        )

        for q in queries:
            logging.info(f"Applying {q}...")
            self._client.sql(q)

        logging.info(f"Feature Set migration finished successfully.")



"""Utils for custom or spark function to generation namedtuple."""

from typing import Callable

from butterfree.constants import DataType


class Function:
    """Define a class Function.

    Like a namedtuple:
        Function = namedtuple("Function", ["function", "data_type"]).

    Attributes:
        func: custom or spark functions, such as avg, std, count.
            For more information check spark functions:
                'https://spark.apache.org/docs/2.3.1/api/python/_modules/pyspark/sql/functions.html'
            For custom functions, look the path:
                'butterfree/transform/transformations/user_defined_functions'.
        data_type: data type for the output columns.
    """

    def __init__(self, func: Callable, data_type: DataType):
        self.func = func
        self.data_type = data_type

    @property
    def func(self) -> Callable:
        """Function to be used in the transformation."""
        return self._func

    @func.setter
    def func(self, value: Callable) -> None:
        """Definitions to be used in the transformation."""
        if not value:
            raise ValueError("Function must not be empty.")
        if not callable(value):
            raise TypeError("Function must be callable.")

        self._func = value

    @property
    def data_type(self) -> DataType:
        """Function to be used in the transformation."""
        return self._data_type

    @data_type.setter
    def data_type(self, value: DataType) -> None:
        """Definitions to be used in the transformation."""
        if not value:
            raise ValueError("DataType must not be empty.")
        if not isinstance(value, DataType):
            raise TypeError("Data type must be DataType.")

        self._data_type = value



"""Write feature set metadata."""

import json
from typing import Any, Dict, List, Optional, Tuple

from mdutils import MdUtils

from butterfree.extract.readers import FileReader, KafkaReader, TableReader
from butterfree.pipelines import FeatureSetPipeline
from butterfree.transform.aggregated_feature_set import AggregatedFeatureSet
from butterfree.transform.transformations import SparkFunctionTransform


class Metadata:
    """Generate metadata for feature set pipeline.

    Attributes:
        pipeline: object processed with feature set pipeline.
        save: bool value with default is False.
            When this value is True, it will generate a file.

    Example:

        >>> pipeline = FeatureSetPipeline()
        >>> metadata = Metadata(pipeline)
        >>> metadata.to_json()

    [
    {
        "feature_set": "feature_set",
        "description": "description",
        "source": [
            {
                "reader": "Table Reader",
                "location": "db.table"
            },
            {
                "reader": "File Reader",
                "location": "path"
            }
        ],
        "sink": [
            {
                "writer": "Historical Feature Store Writer"
            },
            {
                "writer": "Online Feature Store Writer"
            }
        ],
        "features": [
            {
                "column": {
                    "name": "user_id",
                    "data_type": "IntegerType"
                },
                "description": "The user's Main ID or device ID"
            },
            {
                "column": {
                    "name": "timestamp",
                    "data_type": "TimestampType"
                },
                "description": "Time tag for the state of all features."
            },
            {
                "column": {
                    "name":
                        "listing_page_viewed__rent_per_month__avg_over_7_days_fixed_windows",
                    "data_type": "FloatType"
                },
                "description": "Average of something."
            },
            {
                "column": {
                    "name":
                        "listing_page_viewed__rent_per_month__avg_over_2_weeks_fixed_windows",
                    "data_type": "FloatType"
                },
                "description": "Average of something."
            }
        ]
    }
    ]
    """

    def __init__(self, pipeline: FeatureSetPipeline, save: bool = False):
        self.pipeline = pipeline
        self.save = save
        self._name: Optional[str] = None
        self._desc_feature_set: Optional[str] = None
        self._source: List[Tuple[str, str]] = []
        self._sink: List[str] = []
        self._features: List[Tuple[Dict[str, Any], Any]] = []

    def _construct(self) -> "Metadata":
        self._name = self.pipeline.feature_set.name
        self._desc_feature_set = self.pipeline.feature_set.description

        source = []
        for reader in self.pipeline.source.readers:
            if isinstance(reader, TableReader):
                source.append((reader.__name__, f"{reader.database}.{reader.table}"))
            if isinstance(reader, FileReader):
                source.append((reader.__name__, reader.path))
            if isinstance(reader, KafkaReader):
                source.append((reader.__name__, reader.topic))

        self._source = source

        self._sink = [writer.__name__ for writer in self.pipeline.sink.writers]

        desc_feature = [
            feature.description for feature in self.pipeline.feature_set.keys
        ]
        desc_feature.append(self.pipeline.feature_set.timestamp.description)

        for feature in self.pipeline.feature_set.features:
            if feature.transformation:
                windows = (
                    feature.transformation._windows or [None]
                    if isinstance(feature.transformation, SparkFunctionTransform)
                    else (
                        self.pipeline.feature_set._windows or [None]
                        if isinstance(self.pipeline.feature_set, AggregatedFeatureSet)
                        else [None]
                    )
                )
                pivot_values = (
                    self.pipeline.feature_set._pivot_values or [None]
                    if isinstance(self.pipeline.feature_set, AggregatedFeatureSet)
                    else [None]
                )
                desc_feature += (
                    [
                        feature.description
                        for _ in feature.transformation.functions
                        for _ in range(len(pivot_values) * len(windows))
                    ]
                    if isinstance(feature.transformation, SparkFunctionTransform)
                    or isinstance(self.pipeline.feature_set, AggregatedFeatureSet)
                    else [feature.description]
                )
            else:
                desc_feature += [feature.description]

        schema = self.pipeline.feature_set.get_schema()

        self._features = [(column, desc) for column, desc in zip(schema, desc_feature)]

        return self

    def to_json(self) -> Any:
        """Generate json file."""
        params = self._construct()

        lines = [
            {
                "feature_set": params._name,
                "description": params._desc_feature_set,
                "source": [{"reader": r, "location": l} for r, l in params._source],
                "sink": [{"writer": w} for w in params._sink],
                "features": [
                    {
                        "column_name": c["column_name"],
                        "data_type": str(c["type"]),
                        "description": desc,
                    }
                    for c, desc in params._features
                ],
            }
        ]

        if self.save:
            with open(f"{params._name}.json", "w") as outfile:
                json.dump(lines, outfile)
        else:
            return lines

    def to_markdown(self) -> Any:
        """Generate markdown file."""
        params = self._construct()

        markdown = MdUtils(file_name=params._name)
        markdown.new_header(level=1, title=str(params._name).capitalize())
        markdown.new_header(level=2, title="Description")
        markdown.new_paragraph(params._desc_feature_set)
        markdown.new_line()
        markdown.new_header(level=2, title="Feature Set Pipeline")
        markdown.new_header(level=3, title="Source")

        source = ["Reader", "Location"]
        for r, l in params._source:
            source.extend([r, l])

        count_rows = len(source) // 2

        markdown.new_table(columns=2, rows=count_rows, text=source, text_align="center")
        markdown.new_header(level=3, title="Sink")

        sink = ["Writer"]
        for w in params._sink:
            sink.extend([w])

        count_rows = len(sink)

        markdown.new_table(columns=1, rows=count_rows, text=sink, text_align="center")
        markdown.new_header(level=3, title="Features")

        features = ["Column name", "Data type", "Description"]
        for c, desc in params._features:
            features.extend([c["column_name"], str(c["type"]), desc])

        count_rows = len(features) // 3

        markdown.new_table(
            columns=3, rows=count_rows, text=features, text_align="center"
        )

        if self.save:
            markdown.create_md_file()
        else:
            return markdown.file_data_text



"""Hook abstract class entity."""

from abc import ABC, abstractmethod

from pyspark.sql import DataFrame


class Hook(ABC):
    """Definition of a hook function to call on a Dataframe."""

    @abstractmethod
    def run(self, dataframe: DataFrame) -> DataFrame:
        """Run interface for Hook.

        Args:
            dataframe: dataframe to use in the Hook.

        Returns:
            dataframe result from the Hook.
        """



"""Explode json column for dataframes."""
from pyspark.sql.dataframe import DataFrame, StructType
from pyspark.sql.functions import from_json, get_json_object

JSON_TYPE_NAMES = ["array", "struct"]


def explode_json_column(
    df: DataFrame, column: str, json_schema: StructType
) -> DataFrame:
    """Create new columns extracting properties from a JSON column.

    Example:

    >>> from pyspark import SparkContext
    >>> from pyspark.sql import session
    >>> from butterfree.testing.dataframe import create_df_from_collection
    >>> from butterfree.extract.pre_processing import explode_json_column
    >>> from pyspark.sql.types import (
    ...     ArrayType,
    ...     IntegerType,
    ...     StringType,
    ...     StructField,
    ...     StructType,
    ... )
    >>> spark_context = SparkContext.getOrCreate()
    >>> spark_session = session.SparkSession(spark_context)
    >>> data = [{"json_column": '{"a": 123, "b": "abc", "c": "123", "d": [1, 2, 3]}'}]
    >>> df = create_df_from_collection(data, spark_context, spark_session)
    >>> df.collect()

    [Row(json_column='{"a": 123, "b": "abc", "c": "123", "d": [1, 2, 3]}')]

    >>> json_column_schema = StructType(
    ... [
    ...    StructField("a", IntegerType()),
    ...    StructField("b", StringType()),
    ...    StructField("c", IntegerType()),
    ...    StructField("d", ArrayType(IntegerType())),
    ... ]
    >>> explode_json_column(
    ...     df, column='json_column', json_schema=json_column_schema
    ... ).collect()

    [
        Row(
            json_column='{"a": 123, "b": "abc", "c": "123", "d": [1, 2, 3]}',
            a=123,
            b='abc',
            c=123,
            d=[1, 2, 3]
        )
    ]

    Args:
        df: input dataframe with the target JSON column.
        column: column name that is going to be exploded.
        json_schema: expected schema from that JSON column.
            Not all "first layer" fields need to be mapped in the json_schema,
            just the desired columns. If there is any JSON field that is needed
            to be cast to a struct, the declared expected schema (a StructType)
            need to have the exact same schema as the presented record, if don't,
            the value in the resulting column will be null.

    Returns:
        dataframe with the new extracted columns from the JSON column.

    """
    for field in json_schema:
        if field.dataType.typeName() in JSON_TYPE_NAMES:
            df = df.withColumn(
                field.name,
                from_json(
                    get_json_object(df[column], "$.{}".format(field.name)),
                    schema=field.dataType,  # type: ignore
                ),
            )
        else:  # non-collection data types
            df = df.withColumn(
                field.name,
                get_json_object(df[column], "$.{}".format(field.name)).cast(
                    field.dataType
                ),
            )
    return df



"""IncrementalStrategy entity."""

from __future__ import annotations

from pyspark.sql import DataFrame


class IncrementalStrategy:
    """Define an incremental strategy to be used on data sources.

    Entity responsible for defining a column expression that will be used to
    filter the original data source. The purpose is to get only the data related
    to a specific pipeline execution time interval.

    Attributes:
        column: column expression on which incremental filter will be applied.
            The expression need to result on a date or timestamp format, so the
            filter can properly work with the defined upper and lower bounds.
    """

    def __init__(self, column: str = None):
        self.column = column

    def from_milliseconds(self, column_name: str) -> IncrementalStrategy:
        """Create a column expression from ts column defined as milliseconds.

        Args:
            column_name: column name where the filter will be applied.

        Returns:
            `IncrementalStrategy` with the defined column expression.
        """
        return IncrementalStrategy(column=f"from_unixtime({column_name}/ 1000.0)")

    def from_string(self, column_name: str, mask: str = None) -> IncrementalStrategy:
        """Create a column expression from ts column defined as a simple string.

        Args:
            column_name: column name where the filter will be applied.
            mask: mask defining the date/timestamp format on the string.

        Returns:
            `IncrementalStrategy` with the defined column expression.
        """
        return IncrementalStrategy(column=f"to_date({column_name}, '{mask}')")

    def from_year_month_day_partitions(
        self,
        year_column: str = "year",
        month_column: str = "month",
        day_column: str = "day",
    ) -> IncrementalStrategy:
        """Create a column expression from year, month and day partitions.

        Args:
            year_column: column name from the year partition.
            month_column: column name from the month partition.
            day_column: column name from the day partition.

        Returns:
            `IncrementalStrategy` with the defined column expression.
        """
        return IncrementalStrategy(
            column=f"concat(string({year_column}), "
            f"'-', string({month_column}), "
            f"'-', string({day_column}))"
        )

    def get_expression(self, start_date: str = None, end_date: str = None) -> str:
        """Get the incremental filter expression using the defined dates.

        Both arguments can be set to defined a specific date interval, but  it's
        only necessary to set one of the arguments for this method to work.

        Args:
            start_date: date lower bound to use in the filter.
            end_date: date upper bound to use in the filter.

        Returns:
            Filter expression based on defined column and bounds.

        Raises:
            ValuerError: If both arguments, start_date and end_date, are None.
            ValueError: If the column expression was not defined.
        """
        if not self.column:
            raise ValueError("column parameter can't be None")
        if not (start_date or end_date):
            raise ValueError("Both arguments start_date and end_date can't be None.")
        if start_date:
            expression = f"date({self.column}) >= date('{start_date}')"
            if end_date:
                expression += f" and date({self.column}) <= date('{end_date}')"
            return expression
        return f"date({self.column}) <= date('{end_date}')"

    def filter_with_incremental_strategy(
        self, dataframe: DataFrame, start_date: str = None, end_date: str = None
    ) -> DataFrame:
        """Filters the dataframe according to the date boundaries.

        Args:
            dataframe: dataframe that will be filtered.
            start_date: date lower bound to use in the filter.
            end_date: date upper bound to use in the filter.

        Returns:
            Filtered dataframe based on defined time boundaries.
        """
        return (
            dataframe.where(
                self.get_expression(start_date=start_date, end_date=end_date)
            )
            if start_date or end_date
            else dataframe
        )



"""Abstract class for database clients."""
from abc import ABC, abstractmethod
from typing import Any


class AbstractClient(ABC):
    """Abstract base class for database clients."""

    @property
    @abstractmethod
    def conn(self) -> Any:
        """Returns a connection object."""
        pass

    @abstractmethod
    def sql(self, query: str) -> Any:
        """Runs a query.

        Args:
          query: client query.

        Returns:
            Set of records.
        """
        pass



"""Holds common column names, constant through all Butterfree."""

from typing_extensions import Final

TIMESTAMP_COLUMN: Final = "timestamp"
PARTITION_YEAR: Final = "year"
PARTITION_MONTH: Final = "month"
PARTITION_DAY: Final = "day"



"""Spark Function Transform entity."""
from typing import Any, List

from pyspark.sql import DataFrame

from butterfree.transform.transformations.transform_component import TransformComponent
from butterfree.transform.utils import Window
from butterfree.transform.utils.function import Function


class SparkFunctionTransform(TransformComponent):
    """Defines an Spark Function.

    Attributes:
        function: namedtuple with spark function and data type.

    Example:
        It's necessary to declare the function method,
        Any spark and user defined functions are supported.

        >>> from butterfree.transform.transformations import SparkFunctionTransform
        >>> from butterfree.constants.columns import TIMESTAMP_COLUMN
        >>> from butterfree.transform.features import Feature
        >>> from butterfree.transform.utils import Function
        >>> from butterfree.constants import DataType
        >>> from pyspark import SparkContext
        >>> from pyspark.sql import session
        >>> from pyspark.sql.types import TimestampType
        >>> from pyspark.sql import functions
        >>> sc = SparkContext.getOrCreate()
        >>> spark = session.SparkSession(sc)
        >>> df = spark.createDataFrame([(1, "2016-04-11 11:31:11", 200),
        ...                             (1, "2016-04-11 11:44:12", 300),
        ...                             (1, "2016-04-11 11:46:24", 400),
        ...                             (1, "2016-04-11 12:03:21", 500)]
        ...                           ).toDF("id", "timestamp", "feature")
        >>> df = df.withColumn("timestamp", df.timestamp.cast(TimestampType()))
        >>> feature = Feature(
        ...    name="feature",
        ...    description="spark function transform",
        ...    transformation=SparkFunctionTransform(
        ...       functions=[Function(functions.cos, DataType.DOUBLE)],)
        ...)
        >>> feature.transform(df).orderBy("timestamp").show()
        +---+-------------------+-------+--------------------+
        | id|          timestamp|feature|        feature__cos|
        +---+-------------------+-------+--------------------+
        |  1|2016-04-11 11:31:11|    200|  0.4871876750070059|
        |  1|2016-04-11 11:44:12|    300|-0.02209661927868...|
        |  1|2016-04-11 11:46:24|    400|  -0.525296338642536|
        |  1|2016-04-11 12:03:21|    500|  -0.883849273431478|
        +---+-------------------+-------+--------------------+

        We can use this transformation with windows.

        >>> feature_row_windows = Feature(
        ...    name="feature",
        ...    description="spark function transform with windows",
        ...    transformation=SparkFunctionTransform(
        ...       functions=[Function(functions.avg, DataType.DOUBLE)],)
        ...                    .with_window(partition_by="id",
        ...                                 mode="row_windows",
        ...                                 window_definition=["2 events"],
        ...   )
        ...)
        >>> feature_row_windows.transform(df).orderBy("timestamp").show()
        +--------+-----------------------+---------------------------------------+
        |feature | id|          timestamp| feature_avg_over_2_events_row_windows|
        +--------+---+-------------------+--------------------------------------+
        |     200|  1|2016-04-11 11:31:11|                                 200.0|
        |     300|  1|2016-04-11 11:44:12|                                 250.0|
        |     400|  1|2016-04-11 11:46:24|                                 350.0|
        |     500|  1|2016-04-11 12:03:21|                                 450.0|
        +--------+---+-------------------+--------------------------------------+

        It's important to notice that transformation doesn't affect the
        dataframe granularity.

    """

    def __init__(self, functions: List[Function]):
        super().__init__()
        self.functions = functions
        self._windows: List[Any] = []

    def with_window(
        self,
        partition_by: str,
        window_definition: List[str],
        order_by: str = None,
        mode: str = None,
    ) -> "SparkFunctionTransform":
        """Create a list with windows defined."""
        if mode is not None:
            self._windows = [
                Window(
                    partition_by=partition_by,
                    order_by=order_by,
                    mode=mode,
                    window_definition=definition,
                )
                for definition in window_definition
            ]
        return self

    def _get_output_name(self, function: object, window: Window = None) -> str:
        base_name = (
            "__".join([self._parent.name, function.__name__])
            if hasattr(function, "__name__")
            else self._parent.name
        )

        if self._windows and window is not None:
            return "_".join([base_name, window.get_name()])

        return base_name

    @property
    def output_columns(self) -> List[str]:
        """Columns generated by the transformation."""
        output_columns = []
        for function in self.functions:
            if self._windows:
                for window in self._windows:
                    output_columns.append(self._get_output_name(function.func, window))
            else:
                output_columns.append(self._get_output_name(function.func))

        return output_columns

    def transform(self, dataframe: DataFrame) -> DataFrame:
        """Performs a transformation to the feature pipeline.

        Args:
            dataframe: input dataframe.

        Returns:
            Transformed dataframe.
        """
        for function in self.functions:
            if self._windows:
                for window in self._windows:
                    dataframe = dataframe.withColumn(
                        self._get_output_name(function.func, window),
                        function.func(self._parent.from_column or self._parent.name)
                        .over(window.get())
                        .cast(function.data_type.spark),
                    )
            else:
                dataframe = dataframe.withColumn(
                    self._get_output_name(function.func),
                    function.func(self._parent.from_column or self._parent.name).cast(
                        function.data_type.spark
                    ),
                )

        return dataframe



import logging


def __logger(name: str) -> logging.Logger:
    format_ = "%(name)s:%(asctime)-15s:%(levelname)s:< %(message)s >"
    logging.basicConfig(format=format_, level=logging.INFO)
    return logging.getLogger(name)


cli_logger = __logger("butterfree")



"""Holds all transformations to be used by Features.

A transformation must inherit from a TransformComponent and handle data modification,
renaming and cast types using parent's (a Feature) information.
"""

from butterfree.transform.transformations.aggregated_transform import (
    AggregatedTransform,
)
from butterfree.transform.transformations.custom_transform import CustomTransform
from butterfree.transform.transformations.spark_function_transform import (
    SparkFunctionTransform,
)
from butterfree.transform.transformations.sql_expression_transform import (
    SQLExpressionTransform,
)
from butterfree.transform.transformations.stack_transform import StackTransform
from butterfree.transform.transformations.transform_component import TransformComponent

__all__ = [
    "AggregatedTransform",
    "CustomTransform",
    "SparkFunctionTransform",
    "SQLExpressionTransform",
    "StackTransform",
    "TransformComponent",
]



"""FileReader entity."""
from typing import Any, Dict

from pyspark.sql import DataFrame
from pyspark.sql.types import StructType

from butterfree.clients import SparkClient
from butterfree.extract.readers.reader import Reader


class FileReader(Reader):
    """Responsible for get data from files.

    Attributes:
        id: unique string id for register the reader as a view on the metastore.
        path: file location.
        format: can be one of the keys: json, parquet, orc, or csv.
        schema: an optional pyspark.sql.types.StructType for the input schema.
        format_options: additional options required by some formats. Check docs:
            https://spark.apache.org/docs/latest/sql-data-sources-load-save-functions.html#manually-specifying-options

    Example:
        Simple example regarding FileReader class instantiation.

    >>> from butterfree.extract.readers import FileReader
    >>> from butterfree.clients import SparkClient
    >>> from butterfree.extract.pre_processing import filter
    >>> spark_client = SparkClient()
    >>> file_reader = FileReader(
    ...                 id="file_reader_id",
    ...                 path="data_path",
    ...                 format="json"
    ...               )
    >>> df = file_reader.consume(spark_client)

        However, we can define the schema and format_options,
        like header, and provide them to FileReader.

    >>> spark_client = SparkClient()
    >>> schema_csv = StructType([
    ...            StructField("column_a", LongType()),
    ...            StructField("column_b", DoubleType()),
    ...            StructField("coumn_c", StringType())
    ...          ])
    >>> file_reader = FileReader(
    ...                 id="file_reader_id",
    ...                 path="data_path",
    ...                 format="csv",
    ...                 schema=schema_csv,
    ...                 format_options={
    ...                    "header": True
    ...                 }
    ...               )
    >>> df = file_reader.consume(spark_client)


        This last method will use the Spark Client, as default, to read
        the desired file, loading data into a dataframe, according to
        FileReader class arguments.

        It's also possible to define simple transformations within the
        reader's scope:

    >>> file_reader.with_(filter, condition="year = 2019").build(spark_client)

        In this case, however, a temp view will be created, cointaining
        the transformed data.

    """

    __name__ = "File Reader"

    def __init__(
        self,
        id: str,
        path: str,
        format: str,
        schema: StructType = None,
        format_options: Dict[Any, Any] = None,
        stream: bool = False,
    ):
        super().__init__(id)
        if not isinstance(path, str):
            raise ValueError("path must be a string with the file location")
        if not isinstance(format, str):
            raise ValueError("format must be a string with the file type")
        self.path = path
        self.format = format
        self.schema = schema
        self.options = dict(format_options if format_options else {})
        self.stream = stream

    def consume(self, client: SparkClient) -> DataFrame:
        """Extract data from files stored in defined path.

        Try to auto-infer schema if in stream mode and not manually defining a
        schema.

        Args:
            client: client responsible for connecting to Spark session.

        Returns:
            Dataframe with all the files data.

        """
        schema = (
            client.read(format=self.format, path=self.path, **self.options).schema
            if (self.stream and not self.schema)
            else self.schema
        )

        return client.read(
            format=self.format,
            schema=schema,
            stream=self.stream,
            path=self.path,
            **self.options,
        )



"""Holds all feature types to be part of a FeatureSet."""

from butterfree.transform.features.feature import Feature
from butterfree.transform.features.key_feature import KeyFeature
from butterfree.transform.features.timestamp_feature import TimestampFeature

__all__ = ["Feature", "KeyFeature", "TimestampFeature"]



"""Pre Processing Components regarding Readers."""
from butterfree.extract.pre_processing.explode_json_column_transform import (
    explode_json_column,
)
from butterfree.extract.pre_processing.filter_transform import filter
from butterfree.extract.pre_processing.forward_fill_transform import forward_fill
from butterfree.extract.pre_processing.pivot_transform import pivot
from butterfree.extract.pre_processing.replace_transform import replace

__all__ = ["explode_json_column", "filter", "forward_fill", "pivot", "replace"]



from setuptools import find_packages, setup

__package_name__ = "butterfree"
__version__ = "1.2.0.dev4"
__repository_url__ = "https://github.com/quintoandar/butterfree"

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

with open("README.md") as f:
    long_description = f.read()

setup(
    name=__package_name__,
    description="A tool for building feature stores - Transform your raw data "
    "into beautiful features.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="feature store sets ETL",
    version=__version__,
    url=__repository_url__,
    packages=find_packages(
        exclude=(
            "docs",
            "tests",
            "tests.*",
            "pipenv",
            "env",
            "examples",
            "htmlcov",
            ".pytest_cache",
        )
    ),
    license="Copyright",
    author="QuintoAndar",
    install_requires=requirements,
    extras_require={"h3": ["cmake==3.16.3", "h3==3.4.2"]},
    python_requires=">=3.7, <4",
    entry_points={"console_scripts": ["butterfree=butterfree._cli.main:app"]},
    include_package_data=True,
)



"""KeyFeature entity."""

from butterfree.constants.data_type import DataType
from butterfree.transform.features.feature import Feature
from butterfree.transform.transformations import TransformComponent


class KeyFeature(Feature):
    """Defines a KeyFeature.

    A FeatureSet must contain one or more KeyFeatures, which will be used as
    keys when storing the feature set dataframe as tables. The Feature Set may
    validate keys are unique for the latest state of a feature set.

    Attributes:
        name: key name.
            Can be use by the transformation to derive multiple key columns.
        description: brief explanation regarding the key.
        dtype: data type for the output column of this key.
        from_column: original column to build a key.
            Used when there is transformation or the transformation has no
            reference about the column to use for.
        transformation: transformation that will be applied to create this key.
            Keys can be derived by transformations over any data column. Like a
            location hash based on latitude and longitude.

    """

    def __init__(
        self,
        name: str,
        description: str,
        dtype: DataType,
        from_column: str = None,
        transformation: TransformComponent = None,
    ) -> None:
        super(KeyFeature, self).__init__(
            name=name,
            description=description,
            dtype=dtype,
            from_column=from_column,
            transformation=transformation,
        )
